import pathlib
import re

HOME = pathlib.Path.home()
CONFIG_DIR = HOME.joinpath(pathlib.Path(".config"))

MYCFG_CONFIG_DIR = CONFIG_DIR.joinpath(pathlib.Path("mycfg"))
META_FILE = MYCFG_CONFIG_DIR.joinpath(pathlib.Path("meta.json"))
CONFIG_FILE = MYCFG_CONFIG_DIR.joinpath(pathlib.Path("config.yml"))
DOTFILES_SAVE_DIR = MYCFG_CONFIG_DIR.joinpath(pathlib.Path("dotfiles"))
CUSTOM_SCRIPT_DIR = MYCFG_CONFIG_DIR.joinpath(pathlib.Path("scripts"))
BACKUP_DIR = MYCFG_CONFIG_DIR.joinpath(pathlib.Path("backups"))
URL_REGEX = re.compile(
    r"https?://(www\.)?[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_+.~#?&/=]*)")

IGNORE_PATTERNS = ['.git']

DEFAULT_CONFIG_CONTENT = {
    "include": "",
    "units": {},
    "groups": {},
    "environments": {},
    "packages": {},
    "package-managers": {},
}

DEFAULT_GIT_IGNORE = ["meta.json", "backups/"]

PATH_VARIABLES = {
    "$HOME": HOME,
    "$MYCFG_CONFIG": MYCFG_CONFIG_DIR,
}



import pathlib
from setuptools import setup, find_packages


def _file_read(path):
    with open(path, 'r') as f:
        return f.read()


def _file_readlines(path):
    with open(path, 'r') as f:
        return f.readlines()


setup(
    name="mycfg",
    version="1.1.9",
    author="pjones123",
    author_email="pjones-uk@outlook.com",
    description="A basic dotfiles manager",
    long_description=_file_read(pathlib.Path("README.md")),
    long_description_content_type="text/markdown",
    url="https://github.com/pjones123/MyCfg",
    packages=find_packages(),
    install_requires=_file_readlines("requirements.txt"),
    include_package_data=True,
    python_requires=">=3.6",
    keywords=["dotfiles"],
    entry_points={"console_scripts": ["mycfg = mycfg.__main__:main"]},
    classifiers=[
        "Environment :: Console",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Intended Audience :: End Users/Desktop",
    ],
)



import pathlib
import subprocess

import json
import yaml
import re

from mycfg import meta, const
from mycfg.error import PackageManagerError
from mycfg.state import State
from mycfg.parser import parse_environments


def ensure_list(item) -> list:
    if isinstance(item, str):
        return [item]
    return item


def list_diff(a, b):
    return [z for z in a if z not in b]


def sh(cmd):
    return subprocess.call(cmd, shell=True)


def sh_in(_dir, cmd):
    return subprocess.call(cmd, cwd=_dir, shell=True)


def exec_script(script_name):
    return sh_in(const.CUSTOM_SCRIPT_DIR, str(const.CUSTOM_SCRIPT_DIR.joinpath(pathlib.Path(script_name)).resolve()))


def read_file(path: pathlib.Path) -> str:
    with open(path, 'r') as f:
        return f.read()


def write_file(path: pathlib.Path, content: str):
    with open(path, 'w') as f:
        f.write(content)


def get_repo_url(string: str):
    if const.URL_REGEX.match(string):
        return string
    if any(map(string.startswith, ["gh:", "github:"])):
        parts = string.split(":")[1].split("/")
        username = parts[0]
        repo_name = parts[1] if len(parts) > 1 else "dotfiles"
        return f"https://github.com/{username}/{repo_name}"
    else:
        return None


def replace_path_variables(path: str) -> str:
    for k, v in const.PATH_VARIABLES.items():
        path = path.replace(k, v)
    return path


def load_config_file():
    re.match(r"^#include: (.+)((,)(.+))*", "")
    content = re.sub("^#include: (.+)",
                     lambda m: "\n".join(
                         [read_file(pathlib.Path(replace_path_variables(z))) for z in m.group(1).split(",")]),
                     read_file(const.CONFIG_FILE))
    file = yaml.full_load(content)
    State.config = file


def ensure_config_files():
    const.MYCFG_CONFIG_DIR.mkdir(exist_ok=True)
    const.DOTFILES_SAVE_DIR.mkdir(exist_ok=True)
    const.CUSTOM_SCRIPT_DIR.mkdir(exist_ok=True)
    if not const.META_FILE.exists():
        const.META_FILE.write_text(json.dumps(meta.default_meta))
    if not const.CONFIG_FILE.exists():
        const.CONFIG_FILE.write_text(yaml.dump(const.DEFAULT_CONFIG_CONTENT))


def install_pkg(pkg):
    pm = meta.get("package_manager")
    pms = meta.get("package_managers")
    for x in pms:
        if pm is not None:
            break
        pm = x
    if pm is None:
        pm = State.config.get(State.meta.get("environment")).get("default-package-manager", None)
    if pm is None:
        pkg_cfg = State.config.get("packages", {}).get("pkg", {})
        if default := pkg_cfg.get("default", None) is not None:
            sh(default)
        raise PackageManagerError(not_set=True)
    install_cmd = State.config.get("package-managers").get(pm).get("install")
    if install_cmd is None:
        raise PackageManagerError(missing_command="install")
    pkg_cfg = State.config.get("packages", {}).get(pkg, {})
    pkg_name = pkg_cfg.get(pm, pkg_cfg.get("default", None))
    if pkg_name == "none":
        return
    if pkg_name is None:
        raise PackageManagerError(missing_package=pkg)
    sh(f"{install_cmd} {pkg_name}")


def load():
    environment = State.meta.get("environment")
    env_mapping = parse_environments(State.config["environments"])
    env = env_mapping[environment]
    visited_units = []
    units = []
    for group in env.groups:
        units.extend(group.get_all_units(visited_units))
    for unit in units:
        unit.load()


def save():
    environment = State.meta.get("environment")
    env_mapping = parse_environments(State.config["environments"])
    env = env_mapping[environment]
    visited_units = []
    units = []
    for group in env.groups:
        units.extend(group.get_all_units(visited_units))
    for unit in units:
        unit.save()



import pathlib
import shutil

from mycfg import meta, const
from mycfg import lib


class Unit:
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

        self.files = lib.ensure_list(cfg.get("files", []))
        self.preserve_symlinks = cfg.get("preserve-symlinks", "false") == "true"
        self.exclude_files = lib.ensure_list(cfg.get("exclude-files", []))

        self.home_path = pathlib.Path.home()
        self.include_glob = sum([[y for y in self.home_path.glob(z)] for z in self.files], [])
        self.exclude_glob = sum([[y for y in self.home_path.glob(z)] for z in self.exclude_files], [])
        self.glob = lib.list_diff(self.include_glob, self.exclude_glob)

        self.install_commands = lib.ensure_list(cfg.get("install-command", []))
        self.install_scripts = lib.ensure_list(cfg.get("install-script", []))

        self.save_scripts_pre = lib.ensure_list(cfg.get("save-scripts-pre", []))
        self.save_scripts_post = lib.ensure_list(cfg.get("save-scripts-post", []))

        self.load_scripts_pre = lib.ensure_list(cfg.get("save-scripts-pre", []))
        self.load_scripts_post = lib.ensure_list(cfg.get("load-scripts-post", []))

        self.required_packages = lib.ensure_list(cfg.get("requires-packages", []))

    def load(self):
        for pkg in self.required_packages:
            if pkg not in meta.get("installed_packages"):
                lib.install_pkg(pkg)
                meta.append("installed_packages", pkg)
        if self.name not in meta.get("installed_units"):
            for cmd in self.install_commands:
                lib.sh(cmd)
            for script in self.install_scripts:
                lib.exec_script(script)
            meta.append("installed_units", self.name)
        for script in self.load_scripts_pre:
            lib.exec_script(script)
        for file in self.glob:
            src_file = const.DOTFILES_SAVE_DIR.joinpath(file.relative_to(self.home_path))
            if src_file.is_file():
                if not file.exists():
                    file.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(src_file, file)
            elif src_file.is_dir():
                shutil.copytree(src_file, file,
                                symlinks=self.preserve_symlinks, dirs_exist_ok=True, ignore=shutil.ignore_patterns(*const.IGNORE_PATTERNS))
            else:
                print(f"[{src_file}] not found")
        for script in self.load_scripts_post:
            lib.exec_script(script)
        meta.save()

    def save(self):
        for script in self.save_scripts_pre:
            lib.exec_script(script)
        for file in self.glob:
            if file.is_file():
                save_loc = const.DOTFILES_SAVE_DIR.joinpath(file.relative_to(self.home_path))
                if not save_loc.exists():
                    save_loc.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(file, save_loc)
            elif file.is_dir():
                shutil.copytree(file, const.DOTFILES_SAVE_DIR.joinpath(file.relative_to(self.home_path)),
                                symlinks=self.preserve_symlinks, dirs_exist_ok=True, ignore=shutil.ignore_patterns(*const.IGNORE_PATTERNS))
            else:
                print(f"[{file}] not found")
        for script in self.save_scripts_post:
            lib.exec_script(script)
